require "config"
require "guiutils"

local mod_name="tardis"

--overall version of mod
local mod_version="0.1.0"
--data version - last version in which data format changed
local mod_data_version="0.1.0"

local mod_data={
    players_in_tardis={count=0},
    placed=false,
    tardis=nil,
    recent_places={},
    saved_places={},
    gui_states={},
  }

local default_gui_state = {
    open=false,
    tab="recent",
  }

---[[
local function print(...)
  return game.local_player.print(...)
end
--]]

--swap comment to toggle debug printse
--local function debug() end
local function debug(...) end
if debug_mode then
  debug=print
end


local function copy(tab,deep)
  local res={}
  for k,v in pairs(tab) do
    if deep and type(v)=="table" then
      res[k]=copy(v,deep)
    else
      res[k]=v
    end
  end
  return res
end

local function onSave()
  global[mod_name]={data=mod_data, data_version=mod_data_version, version=mod_version}
end

local function onLoad()

  if game.surfaces.tardis==nil then
    game.create_surface("tardis",
      {
        terrain_segmentation="normal",
        water="none",
        autoplace_controls={
          ["iron-ore"]={
             size="none",
            frequency="normal",
            richness="regular",
          },
          ["copper-ore"]={
             size="none",
            frequency="normal",
            richness="regular",
          },
          ["stone"]={
             size="none",
            frequency="normal",
            richness="regular",
          },
          ["crude-oil"]={
             size="none",
            frequency="normal",
            richness="regular",
          },
        },
        seed=0,
        shift={x=0,y=0},
        width=16,
        height=16,
        starting_area="medium",
        peaceful_mode=true,
      })

     game.surfaces.tardis.request_to_generate_chunks({0,0},2)

  end


  if global[mod_name]==nil then
    onSave()
  elseif global[mod_name].data_version~=data_version then
    --migrate
  else
    mod_data=global[mod_name].mod_data
  end

end

local function onChunkGenerated(event)
  if event.surface.name=="tardis" then
    local entities=event.surface.find_entities(event.area)
    for i,entity in pairs(entities) do
      entity.destroy()
    end
    local tiles={}
    local tiles2={}
    local tile={position={x=x,y=y},name="dirt"}

    for y=-8,7 do --event.area.left_top.y,event.area.right_bottom.y do
      for x=-8,7 do --event.area.left_top.x,event.area.right_bottom.x do
        tiles[#tiles+1]={position={x=x,y=y},name="dirt"}
        tiles2[#tiles+1]={position={x=x,y=y},name="concrete"}
      end
    end
    tiles[#tiles+2]={position={-1,8},name="dirt"}
    tiles2[#tiles+2]={position={-1,8},name="concrete"}
    tiles[#tiles+2]={position={0,8},name="dirt"}
    tiles2[#tiles+2]={position={0,8},name="concrete"}
    event.surface.set_tiles(tiles)
    --tile.name="concrete"
    event.surface.set_tiles(tiles2)
  end
end

local function onDrivingChange(event)
  local player=game.players[event.player_index]

  if player.driving and player.vehicle.name=="tardis" then
    player.driving=false
    player.teleport({x=0,y=0},"tardis")
    mod_data.players_in_tardis[player.index]=true
    mod_data.players_in_tardis.count=mod_data.players_in_tardis.count+1
    local desc={type="frame",caption="Tardis HP",direction="vertical",name="tardis_health"}
    local tardis_health=player.gui.left.add(desc)--,player.gui.left)
    desc={type="progressbar",size=1,value=.5,name="tardis_health_bar"}
    tardis_health.add(desc)--createGuiElement(desc,tardis_health)
  end
end

local function onTick(event)
  if event.tick%6==0 then
    if mod_data.players_in_tardis.count>0 then
      local inExit=game.surfaces.tardis.find_entities_filtered{area={{x=-1,y=8},{x=1,y=9}},name="player"}
      for i=1,#inExit do
        local count,found=0,false
        for k,v in pairs(mod_data.players_in_tardis) do
          count=count+1
          if game.players[k].character==inExit[i] then
            game.players[k].teleport(mod_data.tardis.position,mod_data.tardis.surface.name)
            mod_data.players_in_tardis[k]=nil
            mod_data.players_in_tardis.count=mod_data.players_in_tardis.count-1
            game.players[k].gui.left.tardis_health.destroy()
            found=true
            break
          end
        end
      end
    end

    if mod_data.placed and mod_data.tardis.valid then
      local hp=mod_data.tardis.health/1000

      for k,v in pairs(mod_data.players_in_tardis) do
        if k~="count" then
          local player=game.players[k]
          local guileft=player.gui.left
          guileft.tardis_health.tardis_health_bar.value=hp
        end
      end
    end
  end
end

local function onPlaced(event)
  local entity=event.created_entity
  if entity.name=="tardis" then
    if entity.surface.name~="tardis" then
      local oldTardis=mod_data.tardis
      mod_data.tardis=entity
      if mod_data.placed then
        oldTardis.die()
      end
      entity.force=game.forces.neutral
      mod_data.placed=true
      while #mod_data.recent_places>9 do
        table.remove(mod_data.recent_places,10)
      end
      table.insert(mod_data.recent_places,1,{pos=entity.position,surface=entity.surface.name})
    else
      game.players[event.player_index].print("Summoning the tardis inside itself seems like a very bad idea.")
      entity.destroy()
    end

    game.players[event.player_index].cursor_stack.set_stack{name="tardis-key",count=1}

  end
end

local function onDestroyed(event)
  if event.entity==mod_data.tardis then
    --kill all the things! muahaha!
    mod_data.placed=false
    mod_data.tardis=nil
    for k,v in pairs(mod_data.players_in_tardis) do
      if k~="count" then
        game.players[k].print("You died inside the tardis. Sad for you.")
        game.players[k].teleport(event.entity.position,event.entity.surface.name)
      end
    end
  end
end

local function showTardisGui(player_index)
  local player=game.players[player_index]
  local guiState=mod_data.gui_states[player_index]

  --clean up an old one if present
  if player.gui.center.tardis_gui then
    player.gui.center.tardis_gui.destroy()
  end
  if guiState==nil then
    guiState=copy(default_gui_state)
    mod_data.gui_states[player_index]=guiState
  end

  --now on with the new!

  local gui=player.gui.center.add{
      type="frame",
      name="tardis_gui",
      caption="Tardis Control Panel",
      style="tardis_frame_style",
      direction="vertical",
    }

  local main_flow=gui.add{
    type="flow",
    direction="vertical",
    name="tardis_gui_flow",
  }

  local button_flow=main_flow.add{
    type="flow",
    name="tardis_control_button_flow",
    direction="horizontal",
  }
  if guiState.tab=="recent" then
    button_flow.add{
      type="label",
      name="tardis-btn-recent",
      caption="Recent Places",
      style="tardis_selected_tab_style",
    }
  else
    button_flow.add{
      type="button",
      name="tardis-btn-recent",
      caption="Recent Places",
      style="tardis_tab_style",
    }
  end

  button_flow.add{
    type="button",
    name="tardis-btn-saved",
    caption="Saved Places",
    style="tardis_tab_style",
  }


  local list_flow=main_flow.add{
    type="flow",
    name="recent_list",
    direction="vertical"
  }

  if guiState.tab=="recent" then
    for i=1,#mod_data.recent_places do
      local place=mod_data.recent_places[i]
      local item=list_flow.add{
        type="flow",
        name="recent_place_"..i,
        direction="horizontal",
      }
      item.add{
        type="label",
        name="recent_label_"..i,
        caption=place.surface.." @ "..math.floor(place.pos.x)..", "..math.floor(place.pos.y),
        style="tardis_list_label_style",
      }
      item.add{
        type="button",
        name="recent_save_"..i,
        caption="save",
        style="tardis_button_style",
      }
      if i>1 then
        item.add{
          type="button",
          name="recent_go_"..i,
          caption="Go",
          style="tardis_button_style",
        }
      end
    end
  end

  player.gui.center.tardis_gui.tardis_gui_flow.add{
    type="button",
    name="tardis-btn-close",
    caption="Close",
    style="tardis_button_style",
  }

end

local function onGuiClick(event)
  if event.element.name=="tardis-btn-close" then
    game.players[event.player_index].gui.center.tardis_gui.destroy()
  elseif event.element.parent.parent.name=="recent_list" then
    local btn,num=event.element.name:match("^recent_(%a+)_(%d+)$")
    num=tonumber(num)
    if num and num>0 and num<=#mod_data.recent_places then
      if btn=="go" then
        local place=mod_data.recent_places[num]
        local oldTardis=mod_data.tardis
        mod_data.tardis=game.surfaces[place.surface].create_entity{name="tardis",position=place.pos}
        if oldTardis then
          oldTardis.die()
        end
        if num~=1 then
          table.remove(mod_data.recent_places,num)
          table.insert(mod_data.recent_places,1,place)
        end
      end
    end
  end
end


remote.add_interface("tardis",{controls=showTardisGui})

game.on_event(defines.events.on_built_entity,onPlaced)
game.on_event(defines.events.on_entity_died,onDestroyed)

game.on_init(onLoad)
game.on_load(onLoad)

game.on_save(onSave)

game.on_event(defines.events.on_chunk_generated,onChunkGenerated)

game.on_event(defines.events.on_player_driving_changed_state,onDrivingChange)

game.on_event(defines.events.on_tick,onTick)

game.on_event(defines.events.on_gui_click,onGuiClick)